import os, sys, re

import logging
log = logging.getLogger('mbdb')


class Entity(dict):
	def __getattr__(self, key):
		try: return self.__getitem__(key)
		except KeyError:
			if key[0] == '_': raise AttributeError # for data model methods
			else: raise


class Album(Entity):

	cd_suffix = re.compile(r'_cd\d$')

	@classmethod
	def from_path(cls, name, fs_enc='utf-8'):
		year, title = name.decode(fs_enc).split('_', 1)
		try: year = int(year)
		except ValueError: year, title = 0, name
		title = cls.cd_suffix.sub('', title).replace('_', ' ')
		return cls(name=name, year=year, title=title)


class Artist(Entity):

	aggregates_ver = 1

	@classmethod
	def from_path(cls, name, fs_enc='utf-8'):
		return cls(name=name, title=name.decode(fs_enc).replace('_', ' '))

	def aggregate(self, album_list):
		'Generate per-artist aggregate values from associated album releases'
		aggregates_hash = self.aggregates_ver, hash(frozenset(album_list))
		try:
			if self.aggregates_hash != aggregates_hash: raise KeyError
		except KeyError: pass
		else: return

		# Previous years are irrelevant
		self['year_max'] = max(it.imap(op.itemgetter('year'), album_list.itervalues())) if album_list else 0
		# Cache-id for this function results, unique for the same set of albums
		self['aggregates_hash'] = aggregates_hash



rel_drop = map( lambda x: re.compile(x, re.IGNORECASE),
	(r'\s*\([^)]*(disc|cd|live)[^)]*\)', r'cd\d|\bdisc\s*\d+\b') )

def rel_title_strip(title):
	'Drop irrelevant patterns like "(disc 1)"'
	for re in rel_drop: title = re.sub('', title)
	return title



import itertools as it, operator as op, functools as ft
from musicbrainz2.webservice import Query, WebServiceError, ArtistFilter, ArtistIncludes
from musicbrainz2.model import Release
from collections import defaultdict
from difflib import SequenceMatcher
import types, stat, shelve, pickle

try: import pylast
except ImportError: pylast = False

import project.forager.settings as cfg
from project.forager.models import MusicRelease
from project.utils import ext_traceback
from django.db import IntegrityError, transaction

@transaction.commit_manually
def update_releases(*argz, **kwz):
	try: return _update_releases(*argz, **kwz)
	except Exception as err:
		print ext_traceback()
		raise

def _update_releases(music_root, path_list=None, pre_add_hook=None, fs_enc='utf-8'):
	'Get new releases for listed artist paths or every artist path in the music_root'

	cache = shelve.open(cfg.cache_db, writeback=True) if cfg.cache_db else dict()
	seen = pickle.load(open(cfg.seen_db)) if cfg.seen_db and os.path.exists(cfg.seen_db) else set()

	lfm_link = pylast.get_lastfm_network( api_key=cfg.lfm_api_key,
		api_secret=cfg.lfm_api_secret ) if pylast else False

	try: cache_album = cache['album']
	except KeyError: cache_album = cache['album'] = dict()
	try: cache_artist = cache['artist']
	except KeyError: cache_artist = cache['artist'] = dict()

	if path_list is None: path_list = os.listdir(music_root)
	mbz_query = Query()

	for name in path_list:
		name_uc = name.decode(fs_enc, 'replace')
		log.debug(u'Processing {0}...'.format(name_uc))

		path = os.path.join(music_root, name)
		path_stat = os.stat(path)
		if not stat.S_ISDIR(path_stat.st_mode): continue


		# Albums' list
		try: ts, album_list = cache_album[name]
		except KeyError: ts = 0
		if path_stat.st_mtime > ts:
			log.debug(u'Updating album cache for {0}'.format(name_uc))
			album_list = dict( (album.title, album) for album in
				( Album.from_path(album, fs_enc) for album in
					it.ifilter(lambda x: os.path.isdir(os.path.join(path, x)), os.listdir(path)) ) )
			cache_album[name] = path_stat.st_mtime, album_list


		# MBz artist_id
		try: artist = cache_artist[name]
		except KeyError:
			log.debug(u'Updating id cache for {0}'.format(name_uc))
			artist = cache_artist[name] = Artist.from_path(name, fs_enc)
			try:
				query_filter = ArtistFilter(name=artist.title, limit=1)
				artist_mbz = mbz_query.getArtists(query_filter)
			except WebServiceError as err:
				log.error(u'MBz db (getting artist_id): {0}'.format(err))
				artist['id'] = False
			else:
				if not len(artist_mbz): artist['id'] = False
				else:
					artist_mbz = artist_mbz[0].artist
					log.debug( u'Artist name resolved: {0} -> {1}'\
						.format(artist.title, artist_mbz.name) )
					artist['id'] = artist_mbz.id

		if artist.id is False:
			log.info(u'No ID for artist {0}, skipping'.format(name_uc))
			continue


		# Artist by id
		try:
			query_filter = ArtistIncludes(releases=(
				Release.TYPE_OFFICIAL, Release.TYPE_ALBUM ), tags=True)
			artist_mbz = mbz_query.getArtistById(artist.id, query_filter)
		except WebServiceError as err:
			log.error(u'MBz db (getting artist by id): {0}'.format(err))
			continue

		releases = artist_mbz.getReleases()
		if not len(releases):
			log.info(u'No releases found for {0}, skipping'.format(name_uc))
			continue


		# Update associated albums' data, if necessary
		artist.aggregate(album_list)


		# Find/list new releases
		for release in releases:

			# Check against seen_db
			if release.id in seen: continue
			else: seen.add(release.id)

			# Check title against cache_db
			rel_title = rel_title_strip(release.title)
			# Exact match
			if rel_title in album_list: continue
			# Fuzzy match
			matcher = SequenceMatcher(None, rel_title.lower())
			for title in it.imap(op.methodcaller('lower'), album_list):
				matcher.set_seq2(title)
				match = matcher.ratio()
				if match >= cfg.match_threshold: break
			else: match = False
			if match:
				log.debug( u'Found fuzzy-match {0!r} ~ {1!r}, ratio: {2:.2}'\
					.format(matcher.a, matcher.b, match) )
				continue

			# Check if release is clearly obsolete from mbz-db data
			rel_relevance = defaultdict(int)
			rel_events = release.getReleaseEventsAsDict() # usually empty ;(
			for rel_date in rel_events.itervalues(): rel_relevance[int(rel_date.split('-', 1)[0])] += 1
			try: rel_max = max(rel_relevance.iteritems(), key=op.itemgetter(1))
			except ValueError: pass # no events data
			else:
				rel_max = sorted( rel_date for rel_date,rel
					in rel_relevance.iteritems() if rel == rel_max )[0]
				if artist.max_year > rel_max: continue

			rel = MusicRelease(mbdb_id=release.id, artist=artist.title, title=rel_title)
			try: rel.save()
			except IntegrityError: transaction.rollback()
			except:
				log.error(u'Error inserting release into db: {0}'.format(rel))
				raise
			else:
				try:
					# One more "clearly-obsolete" check, based on last.fm data
					rel.update_info(lfm_link) # raises AssertionError on failure to get data
					assert not rel.date or rel.date.year >= artist.year_max
					if pre_add_hook: pre_add_hook(rel) # ...plus the custom hook, JIC
				except AssertionError: transaction.rollback()
				else:
					log.debug(u'Added release: {0}'.format(rel))
					rel.save()
					transaction.commit()

	# Sync
	transaction.commit()
	pickle.dump(seen, open(cfg.seen_db, 'w')) # only synced here, to avoid erroneously skipped releases
	cache.sync()
